class InputValueDuplicateSequenceAddress(ValueError):
    pass


class TaskTemplateMultipleSchemaObjectives(ValueError):
    pass


class TaskTemplateUnexpectedInput(ValueError):
    pass


class TaskTemplateUnexpectedSequenceInput(ValueError):
    pass


class TaskTemplateMultipleInputValues(ValueError):
    pass


class InvalidIdentifier(ValueError):
    pass


class MissingInputs(Exception):
    # TODO: add links to doc pages for common user-exceptions?

    def __init__(self, message, missing_inputs) -> None:
        self.missing_inputs = missing_inputs
        super().__init__(message)


class ExtraInputs(Exception):
    def __init__(self, message, extra_inputs) -> None:
        self.extra_inputs = extra_inputs
        super().__init__(message)


class TaskTemplateInvalidNesting(ValueError):
    pass


class TaskSchemaSpecValidationError(Exception):
    pass


class WorkflowSpecValidationError(Exception):
    pass


class InputSourceValidationError(Exception):
    pass


class EnvironmentSpecValidationError(Exception):
    pass


class ParameterSpecValidationError(Exception):
    pass


class FileSpecValidationError(Exception):
    pass


class DuplicateExecutableError(ValueError):
    pass


class MissingCompatibleActionEnvironment(Exception):
    pass


class MissingActionEnvironment(Exception):
    pass


class FromSpecMissingObjectError(Exception):
    pass


class TaskSchemaMissingParameterError(Exception):
    pass


class ToJSONLikeChildReferenceError(Exception):
    pass


class InvalidInputSourceTaskReference(Exception):
    pass


class WorkflowNotFoundError(Exception):
    pass


class ValuesAlreadyPersistentError(Exception):
    pass


class MalformedParameterPathError(ValueError):
    pass


class UnknownResourceSpecItemError(ValueError):
    pass


class WorkflowParameterMissingError(AttributeError):
    pass


class WorkflowBatchUpdateFailedError(Exception):
    pass


class WorkflowLimitsError(ValueError):
    pass



import numpy as np
import pytest
from hpcflow.api import (
    hpcflow,
    InputValue,
    Parameter,
    SchemaInput,
    Task,
    TaskSchema,
    ValueSequence,
    Workflow,
    WorkflowTemplate,
    InputSource,
    InputSourceType,
    TaskSourceType,
)
from hpcflow.sdk.core.errors import MissingInputs


def test_input_source_class_method_local():
    assert InputSource.local() == InputSource(InputSourceType.LOCAL)


def test_input_source_class_method_default():
    assert InputSource.default() == InputSource(InputSourceType.DEFAULT)


def test_input_source_class_method_task():
    task_ref = 0
    assert InputSource.task(task_ref) == InputSource(
        source_type=InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_class_method_import():
    import_ref = (
        0  # TODO: interface to imports (and so how to reference) is not yet decided
    )
    assert InputSource.import_(import_ref) == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )


def test_input_source_class_method_task_same_default_task_source_type():
    task_ref = 0
    assert (
        InputSource(InputSourceType.TASK, task_ref=task_ref).task_source_type
        == InputSource.task(task_ref=task_ref).task_source_type
    )


def test_input_source_validate_source_type_string_local():
    assert InputSource("local") == InputSource(InputSourceType.LOCAL)


def test_input_source_validate_source_type_string_default():
    assert InputSource("default") == InputSource(InputSourceType.DEFAULT)


def test_input_source_validate_source_type_string_task():
    task_ref = 0
    assert InputSource("task", task_ref=task_ref) == InputSource(
        InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_validate_source_type_string_import():
    import_ref = (
        0  # TODO: interface to imports (and so how to reference) is not yet decided
    )
    assert InputSource("import", import_ref=import_ref) == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )


def test_input_source_validate_source_type_raise_on_unknown_string():
    with pytest.raises(ValueError):
        InputSource("bad_source_type")


def test_input_source_validate_task_source_type_string_any():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="any"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.ANY
    )


def test_input_source_validate_task_source_type_string_input():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="input"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.INPUT
    )


def test_input_source_validate_task_source_type_string_output():
    task_ref = 0
    assert InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type="output"
    ) == InputSource(
        InputSourceType.TASK, task_ref=task_ref, task_source_type=TaskSourceType.OUTPUT
    )


def test_input_source_validate_task_source_type_raise_on_unknown_string():
    task_ref = 0
    with pytest.raises(ValueError):
        InputSource(
            InputSourceType.TASK,
            task_ref=task_ref,
            task_source_type="bad_task_source_type",
        )


def test_input_source_to_string_local():
    assert InputSource.local().to_string() == "local"


def test_input_source_to_string_default():
    assert InputSource.default().to_string() == "default"


def test_input_source_to_string_task_output():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="output").to_string()
        == f"task.{task_ref}.output"
    )


def test_input_source_to_string_task_input():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="input").to_string()
        == f"task.{task_ref}.input"
    )


def test_input_source_to_string_task_any():
    task_ref = 0
    assert (
        InputSource.task(task_ref, task_source_type="any").to_string()
        == f"task.{task_ref}.any"
    )


def test_input_source_to_string_import():
    import_ref = 0
    assert InputSource.import_(import_ref).to_string() == f"import.{import_ref}"


def test_input_source_from_string_local():
    assert InputSource.from_string("local") == InputSource(InputSourceType.LOCAL)


def test_input_source_from_string_default():
    assert InputSource.from_string("default") == InputSource(InputSourceType.DEFAULT)


def test_input_source_from_string_task():
    assert InputSource.from_string("task.0.output") == InputSource(
        InputSourceType.TASK, task_ref=0, task_source_type=TaskSourceType.OUTPUT
    )


def test_input_source_from_string_task_same_default_task_source():
    task_ref = 0
    assert InputSource.from_string(f"task.{task_ref}") == InputSource(
        InputSourceType.TASK, task_ref=task_ref
    )


def test_input_source_from_string_import():
    import_ref = 0
    assert InputSource.from_string(f"import.{import_ref}") == InputSource(
        InputSourceType.IMPORT, import_ref=import_ref
    )


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


@pytest.fixture
def null_config(tmp_path):
    hpcflow.load_config(config_dir=tmp_path)


@pytest.mark.skip(reason="Need to add e.g. parameters of the workflow to the app data.")
def test_specified_sourceable_elements_subset(
    null_config, param_p1, param_p2, param_p3, tmp_path
):

    param_p1 = SchemaInput(param_p1, default_value=1001)
    param_p2 = SchemaInput(param_p2, default_value=np.array([2002, 2003]))
    param_p3 = SchemaInput(param_p3)

    s1 = TaskSchema("ts1", actions=[], inputs=[param_p1], outputs=[param_p3])
    s2 = TaskSchema("ts2", actions=[], inputs=[param_p2, param_p3])

    t1 = Task(
        schemas=s1,
        sequences=[
            ValueSequence("inputs.p1", values=[101, 102], nesting_order=0),
        ],
    )
    t2 = Task(
        schemas=s2,
        inputs=[InputValue(param_p2, 201)],
        sourceable_elements=[0],
        nesting_order={"inputs.p3": 1},
    )

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    wk = Workflow.from_template(wkt, path=tmp_path)

    assert (
        wk.tasks[1].num_elements == 1
        and wk.tasks[1].elements[0].input_sources["inputs.p3"] == "element.0.OUTPUT"
    )


@pytest.mark.skip(reason="Need to add e.g. parameters of the workflow to the app data.")
def test_specified_sourceable_elements_all_available(
    null_config, param_p1, param_p2, param_p3, tmp_path
):

    param_p1 = SchemaInput(param_p1, default_value=1001)
    param_p2 = SchemaInput(param_p2, default_value=np.array([2002, 2003]))
    param_p3 = SchemaInput(param_p3)

    s1 = TaskSchema("ts1", actions=[], inputs=[param_p1], outputs=[param_p3])
    s2 = TaskSchema("ts2", actions=[], inputs=[param_p2, param_p3])

    t1 = Task(
        schemas=s1,
        sequences=[
            ValueSequence("inputs.p1", values=[101, 102], nesting_order=0),
        ],
    )
    t2 = Task(
        schemas=s2,
        inputs=[InputValue(param_p2, 201)],
        sourceable_elements=[0, 1],
        nesting_order={"inputs.p3": 1},
    )

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    wk = Workflow.from_template(wkt, path=tmp_path)

    assert (
        wk.tasks[1].num_elements == 2
        and wk.tasks[1].elements[0].input_sources["inputs.p3"] == "element.0.OUTPUT"
        and wk.tasks[1].elements[1].input_sources["inputs.p3"] == "element.1.OUTPUT"
    )


@pytest.mark.skip(reason="Need to add e.g. parameters of the workflow to the app data.")
def test_no_sourceable_elements_so_raise_missing(
    null_config, param_p1, param_p2, param_p3, tmp_path
):

    param_p1 = SchemaInput(param_p1, default_value=1001)
    param_p2 = SchemaInput(param_p2, default_value=np.array([2002, 2003]))
    param_p3 = SchemaInput(param_p3)

    s1 = TaskSchema("ts1", actions=[], inputs=[param_p1], outputs=[param_p3])
    s2 = TaskSchema("ts2", actions=[], inputs=[param_p2, param_p3])

    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, 101)])
    t2 = Task(
        schemas=s2,
        inputs=[InputValue(param_p2, 201)],
        sourceable_elements=[],
    )

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])

    with pytest.raises(MissingInputs):
        _ = Workflow.from_template(wkt, path=tmp_path)


@pytest.mark.skip(reason="Need to add e.g. parameters of the workflow to the app data.")
def test_no_sourceable_elements_so_default_used(
    null_config, param_p1, param_p2, param_p3, tmp_path
):

    param_p1 = SchemaInput(param_p1, default_value=1001)
    param_p2 = SchemaInput(param_p2, default_value=np.array([2002, 2003]))
    param_p3 = SchemaInput(param_p3, default_value=3001)

    s1 = TaskSchema("ts1", actions=[], inputs=[param_p1], outputs=[param_p3])
    s2 = TaskSchema("ts2", actions=[], inputs=[param_p2, param_p3])

    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, 101)])
    t2 = Task(
        schemas=s2,
        inputs=[InputValue(param_p2, 201)],
        sourceable_elements=[],
    )

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    wk = Workflow.from_template(wkt, path=tmp_path)

    assert wk.tasks[1].elements[0].input_sources["inputs.p3"] == "default"



from hpcflow.config import Config


def get_config_file_contents():
    return Config.config_file_contents


def show_config_file_contents():
    print(get_config_file_contents())


def show_config():
    print(Config.to_string(exclude=["config_file_contents"]))


def check():
    Config.check_data_files()



import pytest
from hpcflow.api import (
    ValueSequence,
    hpcflow,
    Parameter,
    TaskSchema,
    Task,
    WorkflowTemplate,
    Workflow,
)


@pytest.fixture
def null_config(tmp_path):
    hpcflow.load_config(config_dir=tmp_path)


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


@pytest.fixture
def workflow_w1(null_config, tmp_path, param_p1, param_p2):
    s1 = TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p2])
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2])

    t1 = Task(
        schemas=s1,
        sequences=[ValueSequence("inputs.p1", values=[101, 102], nesting_order=1)],
    )
    t2 = Task(schemas=s2, nesting_order={"inputs.p2": 1})

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    return Workflow.from_template(wkt, path=tmp_path)


@pytest.fixture
def workflow_w2(workflow_w1):
    """Add another element set to the second task."""
    workflow_w1.tasks.t2.add_elements(nesting_order={"inputs.p2": 1})
    return workflow_w1



from functools import wraps
import contextlib
import hashlib
import json
import keyword
import os
from pathlib import Path
import random
import re
import socket
import string
from datetime import datetime, timezone
from typing import Mapping, List, Union

import zarr
import numpy.typing as npt
from ruamel.yaml import YAML
import sentry_sdk

from hpcflow.sdk.core.errors import FromSpecMissingObjectError, InvalidIdentifier
from hpcflow.sdk.typing import PathLike


def load_config(func):
    """API function decorator to ensure the configuration has been loaded, and load if not."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.is_config_loaded:
            self.load_config()
        return func(self, *args, **kwargs)

    return wrapper


def make_workflow_id():
    length = 12
    chars = string.ascii_letters + "0123456789"
    return "".join(random.choices(chars, k=length))


def get_time_stamp():
    return datetime.now(timezone.utc).astimezone().strftime("%Y.%m.%d_%H:%M:%S_%z")


def get_duplicate_items(lst):
    """Get a list of all items in an iterable that appear more than once, assuming items
    are hashable.

    Examples
    --------
    >>> get_duplicate_items([1, 1, 2, 3])
    [1]

    >>> get_duplicate_items([1, 2, 3])
    []

    >>> get_duplicate_items([1, 2, 3, 3, 3, 2])
    [2, 3, 2]

    """
    seen = []
    return list(set(x for x in lst if x in seen or seen.append(x)))


def check_valid_py_identifier(name):
    """Check a string is (roughly) a valid Python variable identifier and if so return it
    in lower-case.

    Notes
    -----
    Will be used for:
      - task objective name
      - task method
      - task implementation
      - parameter type
      - parameter name
      - loop name
      - element group name

    """
    trial_name = name[1:].replace("_", "")  # "internal" underscores are allowed
    if (
        not name
        or not (name[0].isalpha() and ((trial_name[1:] or "a").isalnum()))
        or keyword.iskeyword(name)
        or name == "add_object"  # method of `DotAccessObjectList`
    ):
        raise InvalidIdentifier(f"Invalid string for identifier: {name!r}")

    return name.lower()


def group_by_dict_key_values(lst, *keys):
    """Group a list of dicts according to specified equivalent key-values.

    Parameters
    ----------
    lst : list of dict
        The list of dicts to group together.
    keys : tuple
        Dicts that have identical values for all of these keys will be grouped together
        into a sub-list.

    Returns
    -------
    grouped : list of list of dict

    Examples
    --------
    >>> group_by_dict_key_values([{'a': 1}, {'a': 2}, {'a': 1}], 'a')
    [[{'a': 1}, {'a': 1}], [{'a': 2}]]

    """

    grouped = [[lst[0]]]
    for lst_item in lst[1:]:
        for group_idx, group in enumerate(grouped):

            try:
                is_vals_equal = all(lst_item[k] == group[0][k] for k in keys)

            except KeyError:
                # dicts that do not have all `keys` will be in their own group:
                is_vals_equal = False

            if is_vals_equal:
                grouped[group_idx].append(lst_item)
                break

        if not is_vals_equal:
            grouped.append([lst_item])

    return grouped


def get_in_container(cont, path, cast_indices=False):
    cur_data = cont
    for idx, path_comp in enumerate(path):
        if isinstance(cur_data, (list, tuple)):
            if not isinstance(path_comp, int):
                msg = (
                    f"Path component {path_comp!r} must be an integer index "
                    f"since data is a sequence: {cur_data!r}."
                )
                if cast_indices:
                    try:
                        path_comp = int(path_comp)
                    except TypeError:
                        raise TypeError(msg)
                else:
                    raise TypeError(msg)
            cur_data = cur_data[path_comp]
        elif isinstance(cur_data, Mapping):
            cur_data = cur_data[path_comp]
        else:
            raise ValueError(
                f"Data at path {path[:idx]} is not a sequence, but is of type "
                f"{type(cur_data)!r} and so sub-data cannot be extracted."
            )
    return cur_data


def set_in_container(cont, path, value, ensure_path=False):

    if ensure_path:
        num_path = len(path)
        for idx in range(1, num_path):
            try:
                get_in_container(cont, path[:idx])
            except (KeyError, ValueError):
                set_in_container(cont, path[:idx], {}, ensure_path=False)

    sub_data = get_in_container(cont, path[:-1])
    sub_data[path[-1]] = value


def get_relative_path(path1, path2):
    """Get relative path components between two paths.

    Parameters
    ----------
    path1 : tuple of (str or int or float) of length N
    path2 : tuple of (str or int or float) of length less than or equal to N

    Returns
    -------
    relative_path : tuple of (str or int or float)
        The path components in `path1` that are not in `path2`.

    Raises
    ------
    ValueError
        If the two paths do not share a common ancestor of path components, or if `path2`
        is longer than `path1`.

    Notes
    -----
    This function behaves like a simplified `PurePath(*path1).relative_to(PurePath(*path2))`
    from the `pathlib` module, but where path components can include non-strings.

    Examples
    --------
    >>> get_relative_path(('A', 'B', 'C'), ('A',))
    ('B', 'C')

    >>> get_relative_path(('A', 'B'), ('A', 'B'))
    ()

    """

    len_path2 = len(path2)
    msg = f"{path1!r} is not in the subpath of {path2!r}."

    if len(path1) < len_path2:
        raise ValueError(msg)

    for i, j in zip(path1[:len_path2], path2):
        if i != j:
            raise ValueError(msg)

    return path1[len_path2:]


def search_dir_files_by_regex(pattern, group=0, directory="."):
    vals = []
    for i in Path(directory).iterdir():
        match = re.search(pattern, i.name)
        if match:
            match_groups = match.groups()
            if match_groups:
                match = match_groups[group]
                vals.append(match)
    return vals


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class PrettyPrinter(object):
    def __str__(self):
        lines = [self.__class__.__name__ + ":"]
        for key, val in vars(self).items():
            lines += f"{key}: {val}".split("\n")
        return "\n    ".join(lines)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        elif args or kwargs:
            # if existing instance, make the point that new arguments don't do anything!
            raise ValueError(
                f"{cls.__name__!r} is a singleton class and cannot be instantiated with new "
                f"arguments. The positional arguments {args!r} and keyword-arguments "
                f"{kwargs!r} have been ignored."
            )
        return cls._instances[cls]


@contextlib.contextmanager
def sentry_wrap(name, transaction_op=None, span_op=None):
    if not transaction_op:
        transaction_op = name
    if not span_op:
        span_op = name
    try:
        with sentry_sdk.start_transaction(op=transaction_op, name=name):
            with sentry_sdk.start_span(op=span_op) as span:
                yield span
    finally:
        sentry_sdk.flush()  # avoid queue message on stdout


def capitalise_first_letter(chars):
    return chars[0].upper() + chars[1:]


def check_in_object_list(spec_name, spec_pos=1, obj_list_pos=2):
    """Decorator factory for the various `from_spec` class methods that have attributes
    that should be replaced by an object from an object list."""

    def decorator(func):
        @wraps(func)
        def wrap(*args, **kwargs):

            spec = args[spec_pos]
            obj_list = args[obj_list_pos]
            if spec[spec_name] not in obj_list:
                cls_name = args[0].__name__
                raise FromSpecMissingObjectError(
                    f"A {spec_name!r} object required to instantiate the {cls_name!r} "
                    f"object is missing."
                )
            return func(*args, **kwargs)

        return wrap

    return decorator


def read_YAML(loadable_yaml):
    yaml = YAML(typ="safe")
    return yaml.load(loadable_yaml)


def read_YAML_file(path: PathLike):
    return read_YAML(Path(path))


def read_JSON_file(path):
    with Path(path).open("rt") as fh:
        return json.load(fh)


def get_item_repeat_index(lst, distinguish_singular=False, item_callable=None):
    """Get the repeat index for each item in a list.

    Parameters
    ----------
    lst : list
        Must contain hashable items, or hashable objects that are returned via `callable`
        called on each item.
    distinguish_singular : bool, optional
        If True, items that are not repeated will have a repeat index of 0, and items that
        are repeated will have repeat indices starting from 1.
    item_callable : callable, optional
        If specified, comparisons are made on the output of this callable on each item.

    Returns
    -------
    repeat_idx : list of int
        Repeat indices of each item (see `distinguish_singular` for details).

    """

    idx = {}
    for i_idx, item in enumerate(lst):
        if item_callable:
            item = item_callable(item)
        if item not in idx:
            idx[item] = []
        idx[item] += [i_idx]

    rep_idx = [None] * len(lst)
    for k, v in idx.items():
        start = len(v) > 1 if distinguish_singular else 0
        for i_idx, i in enumerate(v, start):
            rep_idx[i] = i_idx

    return rep_idx


def get_process_stamp():
    return "{} {} {}".format(
        datetime.now(),
        socket.gethostname(),
        os.getpid(),
    )


def remove_ansi_escape_sequences(string):
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", string)


def get_md5_hash(obj):
    json_str = json.dumps(obj, sort_keys=True)
    return hashlib.md5(json_str.encode("utf-8")).hexdigest()


def get_nested_indices(idx, size, nest_levels, raise_on_rollover=False):
    """Generate the set of nested indices of length `n` that correspond to a global
    `idx`.

    Examples
    --------
    >>> for i in range(4**2): print(get_nest_index(i, nest_levels=2, size=4))
    [0, 0]
    [0, 1]
    [0, 2]
    [0, 3]
    [1, 0]
    [1, 1]
    [1, 2]
    [1, 3]
    [2, 0]
    [2, 1]
    [2, 2]
    [2, 3]
    [3, 0]
    [3, 1]
    [3, 2]
    [3, 3]

    >>> for i in range(4**3): print(get_nested_indices(i, nest_levels=3, size=4))
    [0, 0, 0]
    [0, 0, 1]
    [0, 0, 2]
    [0, 0, 3]
    [0, 1, 0]
       ...
    [3, 2, 3]
    [3, 3, 0]
    [3, 3, 1]
    [3, 3, 2]
    [3, 3, 3]
    """
    if raise_on_rollover and idx >= size**nest_levels:
        raise ValueError(
            f"`idx` ({idx}) is greater than or equal to  size**nest_levels` "
            f"({size**nest_levels})."
        )

    return [(idx // (size ** (nest_levels - (i + 1)))) % size for i in range(nest_levels)]


def ensure_in(item, lst):
    """Get the index of an item in a list and append the item if it is not in the
    list."""
    # TODO: add tests
    try:
        idx = lst.index(item)
    except ValueError:
        lst.append(item)
        idx = len(lst) - 1
    return idx


def list_to_dict(lst, exclude=None):
    # TODD: test
    exclude = exclude or []
    dct = {k: [] for k in lst[0].keys() if k not in exclude}
    for i in lst:
        for k, v in i.items():
            if k not in exclude:
                dct[k].append(v)

    return dct


def merge_into_zarr_column_array(
    arr: zarr.Array,
    headers: List[str],
    new_arr: Union[npt.NDArray, zarr.Array],
    new_headers: List[str],
) -> List[str]:
    """Append rows into a 2D array whose columns correspond to a list of string
    headers, accounting for different/new headers corresponding to the new array."""

    # TODO: support removal of columns, if a whole column is the fill value?

    if arr.dtype != new_arr.dtype:
        raise ValueError(
            f"New array (dtype {new_arr.dtype}) must have the same dtype as the array "
            f"into which the new array is to be merged (dtype {arr.dtype})."
        )

    if len(headers) != arr.shape[1]:
        raise ValueError(
            "`headers` must be a list of length equal to the number of "
            "columns in `arr`."
        )

    if len(new_headers) != new_arr.shape[1]:
        raise ValueError(
            "`new_headers` must be a list of length equal to the number of "
            "columns in `new_arr`."
        )

    length = arr.shape[0]

    # add new columns:
    add_headers = set(new_headers) - set(headers)
    if add_headers:
        headers += list(add_headers)
        arr.resize(arr.shape[0], len(headers))

    # add new rows:
    arr.resize(arr.shape[0] + new_arr.shape[0], len(headers))

    # get indices of new_headers in new columns:
    head_idx = [headers.index(i) for i in new_headers]

    # fill new row data:
    for idx, i in enumerate(head_idx):
        arr[length:, i] = new_arr[:, idx]

    return headers


def merge_into_headered_zarr_column_array(
    arr: zarr.Array,
    new_arr: zarr.Array,
    header_path: List[str],
):
    """Like `merge_into_zarr_column_array`, but the old and new headers are expected
    at a particular path within the array attributes."""

    attrs = arr.attrs.asdict()
    headers = get_in_container(attrs, header_path)

    new_attrs = new_arr.attrs.asdict()
    new_headers = get_in_container(new_attrs, header_path)

    new_headers = merge_into_zarr_column_array(
        arr=arr,
        headers=headers,
        new_arr=new_arr,
        new_headers=new_headers,
    )
    set_in_container(attrs, header_path, new_headers)
    arr.attrs.put(attrs)


def bisect_slice(selection: slice, len_A: int):
    """Given two sequences (the first of which of known length), get the two slices that
    are equivalent to a given slice if the two sequences were combined."""

    if selection.start < 0 or selection.stop < 0 or selection.step < 0:
        raise NotImplementedError("Can't do negative slices yet.")

    A_idx = selection.indices(len_A)
    B_start = selection.start - len_A
    if len_A != 0 and B_start < 0:
        B_start = B_start % selection.step
    if len_A > selection.stop:
        B_stop = B_start
    else:
        B_stop = selection.stop - len_A
    B_idx = (B_start, B_stop, selection.step)
    A_slice = slice(*A_idx)
    B_slice = slice(*B_idx)

    return A_slice, B_slice



from dataclasses import dataclass
from typing import List, Any

from hpcflow.sdk.core.json_like import JSONLike


@dataclass
class Command(JSONLike):

    _app_attr = "app"

    command: str
    arguments: List[Any] = None
    stdout: str = None
    stderr: str = None
    stdin: str = None

    def __repr__(self) -> str:
        out = []
        if self.command:
            out.append(f"command={self.command!r}")
        if self.arguments:
            out.append(f"arguments={self.arguments!r}")
        if self.stdout:
            out.append(f"stdout={self.stdout!r}")
        if self.stderr:
            out.append(f"stderr={self.stderr!r}")
        if self.stdin:
            out.append(f"stdin={self.stdin!r}")

        return f"{self.__class__.__name__}({', '.join(out)})"


@dataclass
class CommandArgument:
    """
    Attributes
    ----------
    parts : list of any of str, File, Parameter

    """

    parts: List[Any]



from importlib import resources

from valida import Schema


def get_schema(filename):
    """Get a valida `Schema` object from the embedded data directory."""
    with resources.open_text("hpcflow.sdk.data", filename) as fh:
        schema_dat = fh.read()
    schema = Schema.from_yaml(schema_dat)
    return schema



"""API functions, which are dynamically added to the BaseApp class on __init__"""

import importlib

import hpcflow.sdk.scripting
from hpcflow.sdk.core.utils import load_config


@load_config
def make_workflow(app, template_file, dir):
    """Generate a new {name} workflow.

    Parameters
    ----------
    template_file:
        Path to YAML file workflow template.
    dir:
        Directory into which the workflow will be generated.

    Returns
    -------
    Workflow

    """
    app.API_logger.info("make workflow")
    wkt = app.WorkflowTemplate.from_YAML_file(template_file)
    wk = app.Workflow.from_template(wkt, path=dir)
    return wk


@load_config
def submit_workflow(app, template_file, dir):
    """Generate and submit a new {name} workflow.

    Parameters
    ----------
    template_file:
        Path to YAML file workflow template.
    dir:
        Directory into which the workflow will be generated.

    Returns
    -------
    Workflow
    """
    app.API_logger.info("submit workflow")
    wk = app.make_workflow(template_file, dir)
    wk.submit()
    return wk


def run_hpcflow_tests(app, *args):
    """Run hpcflow test suite. This function is only available from derived apps.

    Notes
    -----
    It may not be possible to run hpcflow tests after/before running tests of the derived
    app within the same process, due to caching."""

    from hpcflow.api import hpcflow

    hpcflow.run_tests(*args)


def run_tests(app, *args):
    """Run {name} test suite."""

    try:
        import pytest
    except ModuleNotFoundError:
        raise RuntimeError(f"{app.name} has not been built with testing dependencies.")

    test_args = (app.pytest_args or []) + list(args)
    if app.run_time_info.is_frozen:
        with importlib.resources.path(app.name, "tests") as test_dir:
            pytest.main([str(test_dir)] + test_args)
    else:
        pytest.main(["--pyargs", f"{app.name}"] + test_args)



import copy
from textwrap import dedent
import pytest

from hpcflow.api import (
    Action,
    ActionEnvironment,
    Command,
    Environment,
    FileSpec,
    InputValue,
    OutputFileParser,
    Parameter,
    Task,
    TaskSchema,
    hpcflow,
    WorkflowTemplate,
    Workflow,
)
from hpcflow.sdk.core.errors import (
    MissingInputs,
    WorkflowBatchUpdateFailedError,
    WorkflowNotFoundError,
)


def modify_workflow_metadata_on_disk(workflow):
    """Make a non-sense change to the on-disk metadata."""
    assert workflow.store_format == "zarr"
    wk_md = workflow._store.load_metadata()
    changed_md = copy.deepcopy(wk_md)
    changed_md["new_key"] = "new_value"
    workflow._store._get_root_group(mode="r+").attrs.put(changed_md)


def make_workflow_w1_with_config_kwargs(config_kwargs, path, param_p1, param_p2):
    hpcflow.load_config(**config_kwargs)
    s1 = TaskSchema("ts1", actions=[], inputs=[param_p1], outputs=[param_p2])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, 101)])
    wkt = WorkflowTemplate(name="w1", tasks=[t1])
    return Workflow.from_template(wkt, path=path)


@pytest.fixture
def null_config(tmp_path):
    hpcflow.load_config(config_dir=tmp_path)


@pytest.fixture
def empty_workflow(null_config, tmp_path):
    return Workflow.from_template(WorkflowTemplate(name="w1"), path=tmp_path)


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


@pytest.fixture
def env_1():
    return Environment(name="env_1")


@pytest.fixture
def act_env_1(env_1):
    return ActionEnvironment(env_1)


@pytest.fixture
def act_1(act_env_1):
    return Action(
        commands=[Command("<<parameter:p1>>")],
        environments=[act_env_1],
    )


@pytest.fixture
def act_2(act_env_1):
    return Action(
        commands=[Command("<<parameter:p2>> <<parameter:p3>>")],
        environments=[act_env_1],
    )


@pytest.fixture
def file_spec_fs1():
    return FileSpec(label="file1", name="file1.txt")


@pytest.fixture
def act_3(act_env_1, param_p2, file_spec_fs1):
    return Action(
        commands=[Command("<<parameter:p1>>")],
        output_file_parsers=[
            OutputFileParser(output=param_p2, output_files=[file_spec_fs1]),
        ],
        environments=[act_env_1],
    )


@pytest.fixture
def schema_s1(param_p1, act_1):
    return TaskSchema("ts1", actions=[act_1], inputs=[param_p1])


@pytest.fixture
def schema_s2(param_p2, param_p3, act_2):
    return TaskSchema("ts2", actions=[act_2], inputs=[param_p2, param_p3])


@pytest.fixture
def schema_s3(param_p1, param_p2, act_3):
    return TaskSchema("ts1", actions=[act_3], inputs=[param_p1], outputs=[param_p2])


@pytest.fixture
def workflow_w1(null_config, tmp_path, schema_s3, param_p1):
    t1 = Task(schemas=schema_s3, inputs=[InputValue(param_p1, 101)])
    wkt = WorkflowTemplate(name="w1", tasks=[t1])
    return Workflow.from_template(wkt, path=tmp_path)


def test_make_empty_workflow(empty_workflow):
    assert empty_workflow.path is not None


def test_raise_on_missing_workflow(tmp_path):
    with pytest.raises(WorkflowNotFoundError):
        Workflow(tmp_path)


def test_add_empty_task(empty_workflow, schema_s1):
    t1 = Task(schemas=schema_s1)
    wk_t1 = empty_workflow._add_empty_task(t1)
    assert len(empty_workflow.tasks) == 1 and wk_t1.index == 0 and wk_t1.name == "ts1"


def test_raise_on_missing_inputs_add_first_task(empty_workflow, schema_s1, param_p1):
    t1 = Task(schemas=schema_s1)
    with pytest.raises(MissingInputs) as exc_info:
        empty_workflow.add_task(t1)

    assert exc_info.value.missing_inputs == [param_p1.typ]


def test_raise_on_missing_inputs_add_second_task(workflow_w1, schema_s2, param_p3):
    t2 = Task(schemas=schema_s2)
    with pytest.raises(MissingInputs) as exc_info:
        workflow_w1.add_task(t2)

    assert exc_info.value.missing_inputs == [param_p3.typ]  # p2 comes from existing task


@pytest.mark.skip(reason="TODO: Not implemented.")
def test_new_workflow_deleted_on_creation_failure():
    pass


def test_WorkflowTemplate_from_YAML_string(null_config):
    wkt_yml = dedent(
        """
        name: simple_workflow

        tasks:
        - schemas: [dummy_task_1]
          element_sets:
          - inputs:
              p2: 201
              p5: 501
            sequences:
              - path: inputs.p1
                nesting_order: 0
                values: [101, 102]
    """
    )
    WorkflowTemplate.from_YAML_string(wkt_yml)


def test_WorkflowTemplate_from_YAML_string_without_element_sets(null_config):
    wkt_yml = dedent(
        """
        name: simple_workflow

        tasks:
        - schemas: [dummy_task_1]
          inputs:
            p2: 201
            p5: 501
          sequences:
            - path: inputs.p1
              nesting_order: 0
              values: [101, 102]
    """
    )
    WorkflowTemplate.from_YAML_string(wkt_yml)


def test_WorkflowTemplate_from_YAML_string_with_and_without_element_sets_equivalence(
    null_config,
):
    wkt_yml_1 = dedent(
        """
        name: simple_workflow

        tasks:
        - schemas: [dummy_task_1]
          element_sets:
            - inputs:
                p2: 201
                p5: 501
              sequences:
                - path: inputs.p1
                  nesting_order: 0
                  values: [101, 102]
    """
    )
    wkt_yml_2 = dedent(
        """
        name: simple_workflow

        tasks:
        - schemas: [dummy_task_1]
          inputs:
            p2: 201
            p5: 501
          sequences:
            - path: inputs.p1
              nesting_order: 0
              values: [101, 102]
    """
    )
    wkt_1 = WorkflowTemplate.from_YAML_string(wkt_yml_1)
    wkt_2 = WorkflowTemplate.from_YAML_string(wkt_yml_2)
    assert wkt_1 == wkt_2


def test_store_has_pending_during_add_task(workflow_w1, schema_s2, param_p3):
    t2 = Task(schemas=schema_s2, inputs=[InputValue(param_p3, 301)])
    with workflow_w1.batch_update():
        workflow_w1.add_task(t2)
        assert workflow_w1._store.has_pending


def test_empty_batch_update_does_nothing(workflow_w1):
    with workflow_w1.batch_update():
        assert not workflow_w1._store.has_pending


def test_is_modified_on_disk_when_metadata_changed(workflow_w1):
    # this is ZarrPersistentStore-specific; might want to consider a refactor later
    with workflow_w1._store.cached_load():
        modify_workflow_metadata_on_disk(workflow_w1)
        assert workflow_w1._store.is_modified_on_disk()


def test_batch_update_abort_if_modified_on_disk(workflow_w1, schema_s2, param_p3):
    t2 = Task(schemas=schema_s2, inputs=[InputValue(param_p3, 301)])
    with pytest.raises(WorkflowBatchUpdateFailedError):
        with workflow_w1._store.cached_load():
            with workflow_w1.batch_update():
                workflow_w1.add_task(t2)
                modify_workflow_metadata_on_disk(workflow_w1)



from __future__ import annotations
from abc import ABC, abstractmethod
from contextlib import contextmanager
import copy
import random
import string
from typing import Any, Dict, Generator, Iterator, List, Optional, Union
from pathlib import Path

from reretry import retry
from hpcflow.sdk.core.errors import WorkflowNotFoundError
from hpcflow.sdk.core.parameters import ParameterValue

from hpcflow.sdk.core.utils import get_in_container, get_relative_path, set_in_container
from hpcflow.sdk.typing import PathLike

PRIMITIVES = (
    int,
    float,
    str,
    type(None),
)


def dropbox_retry_fail(err: Exception) -> None:
    # TODO: this should log instead of printing!
    print("retrying...")


# TODO: maybe this is only an issue on Windows?
dropbox_permission_err_retry = retry(
    (PermissionError, OSError),
    tries=10,
    delay=1,
    backoff=2,
    fail_callback=dropbox_retry_fail,
)


class PersistentStore(ABC):

    _parameter_encoders = {}
    _parameter_decoders = {}

    def __init__(self, workflow: Workflow) -> None:
        self._workflow = workflow
        self._pending = self._get_pending_dct()
        if not self.exists():
            raise WorkflowNotFoundError(f"No workflow found at path: {self.path}")

    @property
    def workflow(self) -> Workflow:
        return self._workflow

    @property
    def path(self) -> Path:
        return self.workflow.path

    @property
    def has_pending(self) -> bool:
        """Returns True if there are pending changes that are not yet committed."""
        return any(bool(v) for v in self._pending.values())

    def _get_pending_dct(self) -> Dict:
        return {
            "tasks": {},  # keys are new task indices
            "template_tasks": {},  # keys are new task indices
            "template_components": {},
            "element_sets": {},  # keys are task indices
            "elements": {},  # keys are (task index, task insert ID)
            "parameter_data": {},  # keys are parameter indices
            "parameter_sources": {},  # keys are parameter indices
            "remove_replaced_file_record": False,
        }

    def reject_pending(self) -> None:
        self.clear_pending()

    def clear_pending(self) -> None:
        self._pending = self._get_pending_dct()

    def save(self) -> None:
        if not self.workflow._in_batch_mode:
            self.commit_pending()

    @contextmanager
    def cached_load(self) -> Iterator[None]:
        """Override this if a more performant implementation, is possible.

        For example, in a JSON persistent store, we need to load the whole document from
        disk to read anything from it, so we can temporarily cache the document if we know
        we will be making multiple reads."""

        yield

    def get_task_elements_islice(
        self,
        task_idx: int,
        task_insert_ID: int,
        selection: Union[int, slice],
    ) -> Iterator[Dict]:
        """Override this for a more performant implementation."""
        for idx in range(selection.start, selection.stop, selection.step):
            yield self.get_task_elements(
                task_idx, task_insert_ID, slice(idx, idx + 1, 1)
            )[0]

    def delete(self) -> None:
        """Delete the persistent workflow."""
        confirm = input(
            f"Permanently delete the workflow at path {self.workflow.path}; "
            f"[y]es or [n]o?"
        )
        if confirm.strip().lower() == "y":
            self.delete_no_confirm()

    def _merge_pending_template_components(self, template_components: Dict) -> None:
        # assumes we have already checked for duplicates when adding to pending:
        for name, dat in self._pending["template_components"].items():
            if name not in template_components:
                template_components[name] = {}
            for k, v in dat.items():
                template_components[name][k] = v

    @classmethod
    def _rename_existing(cls, path):
        # rename existing file so we can restore it if workflow creation fails:
        temp_ext = "".join(random.choices(string.ascii_letters, k=8))
        replaced_file = path.with_suffix(f"{path.suffix}.{temp_ext}")
        path.rename(replaced_file)
        return replaced_file

    def get_template_components(self) -> Dict:
        """Get all template components, including pending."""
        tc = self._get_persistent_template_components()
        if self._pending["template_components"]:
            tc = copy.deepcopy(tc)
            self._merge_pending_template_components(tc)
        return tc

    def add_template_components(self, template_components: Dict) -> None:

        ptc = self._get_persistent_template_components()
        pending = self._pending["template_components"]

        for name, dat in template_components.items():

            if name in ptc and name in pending:
                for hash, dat_i in dat.items():
                    if hash not in ptc[name] and hash not in pending[name]:
                        pending[name][hash] = dat_i

            elif name in pending:
                for hash, dat_i in dat.items():
                    if hash not in pending[name]:
                        pending[name][hash] = dat_i

            else:
                pending[name] = dat

        self.save()

    def add_empty_task(self, task_idx: int, task_js: Dict) -> None:
        self._pending["template_tasks"][task_idx] = task_js
        self._pending["tasks"][task_idx] = {"elements": []}
        self.save()

    def add_element_set(self, task_idx: int, element_set_js: Dict) -> None:
        if task_idx not in self._pending["element_sets"]:
            self._pending["element_sets"][task_idx] = []
        self._pending["element_sets"][task_idx].append(element_set_js)
        self.save()

    def add_elements(
        self,
        task_idx: int,
        task_insert_ID: int,
        elements: List[Dict],
    ) -> None:
        key = (task_idx, task_insert_ID)
        if key not in self._pending["elements"]:
            self._pending["elements"][key] = []
        self._pending["elements"][key].extend(elements)
        self.save()

    def add_parameter_data(self, data: Any, source: Dict) -> int:
        return self._add_parameter_data({"data": data}, source)

    def add_unset_parameter_data(self, source: Dict) -> int:
        return self._add_parameter_data(None, source)

    def _encode_parameter_data(
        self,
        obj: Any,
        path: List = None,
        type_lookup: Optional[Dict] = None,
        **kwargs,
    ) -> Any:

        path = path or []
        if type_lookup is None:
            type_lookup = {
                "tuples": [],
                "sets": [],
                **{k: [] for k in self._parameter_decoders.keys()},
            }

        if len(path) > 50:
            raise RuntimeError("I'm in too deep!")

        if isinstance(obj, ParameterValue):
            encoded = self._encode_parameter_data(
                obj=obj.to_dict(),
                path=path,
                type_lookup=type_lookup,
                **kwargs,
            )
            data, type_lookup = encoded["data"], encoded["type_lookup"]

        elif isinstance(obj, (list, tuple, set)):
            data = []
            for idx, item in enumerate(obj):
                encoded = self._encode_parameter_data(
                    obj=item,
                    path=path + [idx],
                    type_lookup=type_lookup,
                    **kwargs,
                )
                item, type_lookup = encoded["data"], encoded["type_lookup"]
                data.append(item)

            if isinstance(obj, tuple):
                type_lookup["tuples"].append(path)

            elif isinstance(obj, set):
                type_lookup["sets"].append(path)

        elif isinstance(obj, dict):
            data = {}
            for dct_key, dct_val in obj.items():
                encoded = self._encode_parameter_data(
                    obj=dct_val,
                    path=path + [dct_key],
                    type_lookup=type_lookup,
                    **kwargs,
                )
                dct_val, type_lookup = encoded["data"], encoded["type_lookup"]
                data[dct_key] = dct_val

        elif isinstance(obj, PRIMITIVES):
            data = obj

        elif type(obj) in self._parameter_encoders:
            data = self._parameter_encoders[type(obj)](
                obj=obj,
                path=path,
                type_lookup=type_lookup,
                **kwargs,
            )

        else:
            raise ValueError(
                f"Parameter data with type {type(obj)} cannot be serialised into a "
                f"{self.__class__.__name__}: {obj}."
            )

        return {"data": data, "type_lookup": type_lookup}

    def _decode_parameter_data(
        self,
        data: Union[None, Dict],
        path: Optional[List[str]] = None,
        **kwargs,
    ) -> Any:

        if data is None:
            return None

        path = path or []

        obj = get_in_container(data["data"], path)

        for tuple_path in data["type_lookup"]["tuples"]:
            try:
                rel_path = get_relative_path(tuple_path, path)
            except ValueError:
                continue
            if rel_path:
                set_in_container(obj, rel_path, tuple(get_in_container(obj, rel_path)))
            else:
                obj = tuple(obj)

        for set_path in data["type_lookup"]["sets"]:
            try:
                rel_path = get_relative_path(set_path, path)
            except ValueError:
                continue
            if rel_path:
                set_in_container(obj, rel_path, set(get_in_container(obj, rel_path)))
            else:
                obj = set(obj)

        for data_type in self._parameter_decoders:
            obj = self._parameter_decoders[data_type](
                obj=obj,
                type_lookup=data["type_lookup"],
                path=path,
                **kwargs,
            )

        return obj

    @classmethod
    @abstractmethod
    def write_empty_workflow(
        cls,
        template_js: Dict,
        template_components_js: Dict,
        path: Path,
        overwrite: bool,
    ) -> None:
        pass

    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def commit_pending(self) -> None:
        pass

    @abstractmethod
    def _get_persistent_template_components(self) -> Dict:
        """Get currently persistent template components, excluding pending."""

    @abstractmethod
    def get_template(self) -> Dict:
        pass

    @abstractmethod
    def get_all_tasks_metadata(self) -> List[Dict]:
        pass

    @abstractmethod
    def get_task_elements(
        self,
        task_idx: int,
        task_insert_ID: int,
        selection: slice,
    ) -> List:
        pass

    @abstractmethod
    def _add_parameter_data(self, data: Any, source: Dict) -> int:
        pass

    @abstractmethod
    def get_parameter_data(self, index: int) -> Any:
        pass

    @abstractmethod
    def get_parameter_source(self, index: int) -> Dict:
        pass

    @abstractmethod
    def get_all_parameter_data(self) -> Dict[int, Any]:
        pass

    @abstractmethod
    def is_parameter_set(self, index: int) -> bool:
        pass

    @abstractmethod
    def set_parameter(self, index: int, data: Any) -> None:
        """Set the value of a pre-allocated parameter."""
        pass

    @abstractmethod
    def check_parameters_exist(
        self, indices: Union[int, List[int]]
    ) -> Union[bool, List[bool]]:
        pass

    @abstractmethod
    def delete_no_confirm(self) -> None:
        """Permanently delete the workflow data with no confirmation."""

    @abstractmethod
    def remove_replaced_file(self) -> None:
        pass

    @abstractmethod
    def reinstate_replaced_file(self) -> None:
        pass

    @abstractmethod
    def copy(self, path: PathLike = None) -> None:
        """Make a copy of the store."""
        pass

    @abstractmethod
    def is_modified_on_disk(self) -> bool:
        """Check if the workflow (metadata) has been modified on disk since initial
        load (this is bad)."""
        pass

    @abstractmethod
    def get_num_added_tasks(self) -> int:
        """Get the total number of tasks ever added to the workflow, regardless of whether
        any of those tasks were subsequently removed from the workflow."""



from contextlib import contextmanager
import copy
from dataclasses import dataclass, field
from typing import List, Optional, Union

from .actions import Action
from .json_like import ChildObjectSpec, JSONLike
from .parameters import (
    Parameter,
    ParameterPropagationMode,
    SchemaInput,
    SchemaOutput,
    SchemaParameter,
)
from .utils import check_valid_py_identifier


@dataclass
class TaskObjective(JSONLike):

    _child_objects = (
        ChildObjectSpec(
            name="name",
            is_single_attribute=True,
        ),
    )

    name: str

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


class TaskSchema(JSONLike):

    _validation_schema = "task_schema_spec_schema.yaml"
    _hash_value = None
    _validate_actions = True

    _child_objects = (
        ChildObjectSpec(name="objective", class_name="TaskObjective"),
        ChildObjectSpec(
            name="inputs",
            class_name="SchemaInput",
            is_multiple=True,
            parent_ref="_task_schema",
        ),
        ChildObjectSpec(name="outputs", class_name="SchemaOutput", is_multiple=True),
        ChildObjectSpec(name="actions", class_name="Action", is_multiple=True),
    )

    def __init__(
        self,
        objective: Union[TaskObjective, str],
        actions: List[Action],
        method: Optional[str] = None,
        implementation: Optional[str] = None,
        inputs: Optional[List[Union[Parameter, SchemaInput]]] = None,
        outputs: Optional[List[Union[Parameter, SchemaOutput]]] = None,
        version: Optional[str] = None,
        _hash_value: Optional[str] = None,
    ):
        self.objective = objective
        self.actions = actions
        self.method = method
        self.implementation = implementation
        self.inputs = inputs or []
        self.outputs = outputs or []
        self._hash_value = _hash_value

        self._validate()
        self.actions = self._expand_actions()
        self.version = version
        self._task_template = None  # assigned by parent Task

        self._set_parent_refs()

        # if version is not None:  # TODO: this seems fragile
        #     self.assign_versions(
        #         version=version,
        #         app_data_obj_list=self.app.task_schemas
        #         if self.app.is_data_files_loaded
        #         else [],
        #     )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"objective={self.objective.name!r}, "
            f"input_types={self.input_types!r}, "
            f"output_types={self.output_types!r}"
            f")"
        )

    def __eq__(self, other):
        if type(other) is not self.__class__:
            return False
        if (
            self.objective == other.objective
            and self.actions == other.actions
            and self.method == other.method
            and self.implementation == other.implementation
            and self.inputs == other.inputs
            and self.outputs == other.outputs
            and self.version == other.version
            and self._hash_value == other._hash_value
        ):
            return True
        return False

    def __deepcopy__(self, memo):
        kwargs = self.to_dict()
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._task_template = self._task_template
        return obj

    @classmethod
    @contextmanager
    def ignore_invalid_actions(cls):
        try:
            cls._validate_actions = False
            yield
        finally:
            cls._validate_actions = True

    def _validate(self):

        if isinstance(self.objective, str):
            self.objective = self.app.TaskObjective(self.objective)

        if self.method:
            self.method = check_valid_py_identifier(self.method)
        if self.implementation:
            self.implementation = check_valid_py_identifier(self.implementation)

        # coerce Parameters to SchemaInputs
        for idx, i in enumerate(self.inputs):
            if isinstance(i, Parameter):
                self.inputs[idx] = self.app.SchemaInput(i)

        # coerce Parameters to SchemaOutputs
        for idx, i in enumerate(self.outputs):
            if isinstance(i, Parameter):
                self.outputs[idx] = self.app.SchemaOutput(i)
            elif isinstance(i, SchemaInput):
                self.outputs[idx] = self.app.SchemaOutput(i.parameter)

        # check action input/outputs
        if self._validate_actions:
            all_outs = []
            extra_ins = set(self.input_types)

            act_ins_lst = [act.get_input_types() for act in self.actions]
            act_outs_lst = [act.get_output_types() for act in self.actions]

            schema_ins = set(self.input_types)
            schema_outs = set(self.output_types)

            all_act_ins = set(j for i in act_ins_lst for j in i)
            all_act_outs = set(j for i in act_outs_lst for j in i)

            non_schema_act_ins = all_act_ins - schema_ins
            non_schema_act_outs = set(all_act_outs - schema_outs)

            extra_act_outs = non_schema_act_outs
            seen_act_outs = []
            for act_idx in range(len(self.actions)):
                for act_in in [
                    i for i in act_ins_lst[act_idx] if i in non_schema_act_ins
                ]:
                    if act_in not in seen_act_outs:
                        raise ValueError(
                            f"Action {act_idx} input {act_in!r} of schema {self.name!r} "
                            f"is not a schema input, but nor is it an action output from "
                            f"a preceding action."
                        )
                seen_act_outs += [
                    i for i in act_outs_lst[act_idx] if i not in seen_act_outs
                ]
                extra_act_outs = extra_act_outs - set(act_ins_lst[act_idx])
                act_inputs = set(act_ins_lst[act_idx])
                act_outputs = set(act_outs_lst[act_idx])
                extra_ins = extra_ins - act_inputs
                all_outs.extend(list(act_outputs))

            if extra_act_outs:
                raise ValueError(
                    f"The following action outputs of schema {self.name!r} are not schema"
                    f" outputs, but nor are they consumed by subsequent actions as "
                    f"action inputs: {tuple(extra_act_outs)!r}."
                )

            if extra_ins:
                # i.e. are all schema inputs "consumed" by an action?
                raise ValueError(
                    f"Schema {self.name!r} inputs {tuple(extra_ins)!r} are not used by "
                    f"any actions."
                )

            missing_outs = set(self.output_types) - set(all_outs)
            if missing_outs:
                raise ValueError(
                    f"Schema {self.name!r} outputs {tuple(missing_outs)!r} are not "
                    f"generated by any actions."
                )

    def _expand_actions(self):
        """Create new actions for input file generators and output parsers in existing
        actions."""
        return [j for i in self.actions for j in i.expand()]

    def make_persistent(self, workflow, source):
        new_refs = []
        for input_i in self.inputs:
            if input_i.default_value is not None:
                _, dat_ref, is_new = input_i.default_value.make_persistent(
                    workflow, source
                )
                new_refs.extend(dat_ref) if is_new else None
        return new_refs

    @property
    def name(self):
        out = (
            f"{self.objective.name}"
            f"{f'_{self.method}' if self.method else ''}"
            f"{f'_{self.implementation}' if self.implementation else ''}"
        )
        return out

    @property
    def input_types(self):
        return tuple(i.typ for i in self.inputs)

    @property
    def output_types(self):
        return tuple(i.typ for i in self.outputs)

    @property
    def provides_parameters(self):
        return tuple(
            i
            for i in self.inputs + self.outputs
            if i.propagation_mode != ParameterPropagationMode.NEVER
        )

    @property
    def task_template(self):
        return self._task_template

    @classmethod
    def get_by_key(cls, key):
        """Get a config-loaded task schema from a key."""
        return cls.app.task_schemas.get(key)

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the schema's actions."""
        out = {"input_file_writers": [], "commands": []}
        for act_idx, action in enumerate(self.actions):
            deps = action.get_parameter_dependence(parameter)
            for key in out:
                out[key].extend((act_idx, i) for i in deps[key])
        return out

    def get_key(self):
        return (str(self.objective), self.method, self.implementation)
